# -*- coding: utf-8 -*-

'''
Created on 20/02/2011

@author: vbmendes
'''

from django import template
from django.test import TestCase

from easytags.node import EasyNode, EasyAsNode


class ParserTests(TestCase):

    def test_environment(self):
        """
            Just make sure everything is set up correctly.
        """
        self.assertTrue(True)

    def test_parse_tag_with_args(self):
        """
            Tests if the parser recognizes one tag and parses its args
        """
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "arg1" "arg2"')
        self.assertEquals(
            {'args': ('"arg1"', '"arg2"'), 'kwargs': {}},
            EasyNode.parse_to_args_kwargs(parser, token)
        )
    
    def test_parse_tag_with_kwargs(self):
        """
            Tests if the parser recognizes one tag and parses its kwargs
        """
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name kwarg1="1" kwarg2="2"')
        self.assertEquals(
            {'args': (), 'kwargs': {'kwarg1': '"1"', 'kwarg2': '"2"'}},
            EasyNode.parse_to_args_kwargs(parser, token)
        )
    
    def test_parse_tag_with_args_and_kwargs(self):
        """
            Tests if the parser recognizes one tag and parses its args and kwargs
        """
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "arg1" kwarg1="1"')
        self.assertEquals(
            {'args': ('"arg1"',), 'kwargs': {'kwarg1': '"1"'}},
            EasyNode.parse_to_args_kwargs(parser, token)
        )

    def test_parse_tag_with_variable_arg(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name argvariable')
        self.assertEquals(
            {'args': ('argvariable',), 'kwargs': {}},
            EasyNode.parse_to_args_kwargs(parser, token)
        )
    
    def test_parse_tag_with_equals_in_arg_value(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "a=1"')
        self.assertEquals(
            {'args': ('"a=1"',), 'kwargs': {}},
            EasyNode.parse_to_args_kwargs(parser, token)
        )
    
    def test_parse_tag_with_equals_in_kwarg_value(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name kwarg1="a=1"')
        self.assertEquals(
            {'args': (), 'kwargs': {'kwarg1': '"a=1"'}},
            EasyNode.parse_to_args_kwargs(parser, token)
        )

    def test_parse_tag_special_symbol_in_arg_value(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, u'tag_name "será?"')
        self.assertEquals(
            {'args': (u'"será?"',), 'kwargs': {}},
            EasyNode.parse_to_args_kwargs(parser, token)
        )

    def test_parse_tag_special_symbol_in_kwarg_value(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, u'tag_name kwarg1="será?"')
        self.assertEquals(
            {'args': (), 'kwargs': {'kwarg1': u'"será?"'}},
            EasyNode.parse_to_args_kwargs(parser, token)
        )

    def test_parse_tag_with_args_after_kwargs_raises_exception(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, u'tag_name kwarg1="será?" my_arg')
        self.assertRaises(template.TemplateSyntaxError,
            EasyNode.parse_to_args_kwargs, parser, token
        )
    
    def test_parse_as_tag_with_args(self):
        """
            Tests if the parser recognizes one tag and parses its args even when using EasyAsNode
        """
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "arg1" "arg2"')
        self.assertEquals(
            {'args': ('"arg1"', '"arg2"'), 'kwargs': {}, 'varname': None},
            EasyAsNode.parse_to_args_kwargs(parser, token)
        )

    def test_parse_as_tag_with_args_and_as_parameter(self):
        """
            Tests if the parser recognizes one tag and parses its args and as parameter
        """
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "arg1" "arg2" as varname')
        self.assertEquals(
            {'args': ('"arg1"', '"arg2"'), 'kwargs': {}, 'varname': 'varname'},
            EasyAsNode.parse_to_args_kwargs(parser, token)
        )



# -*- coding: utf-8 -*-

'''
Created on 01/03/2011

@author: vbmendes
'''

from django.template import Library

from node import EasyNode, EasyAsNode

class EasyLibrary(Library):

    @classmethod
    def _get_name_and_renderer(cls, name, renderer):
        if not renderer:
            renderer = name
            name = renderer.__name__
        return name, renderer
    
    def easytag(self, name = None, renderer = None):
        return self._handle_decorator(EasyNode, name, renderer)
        
    def easyastag(self, name = None, renderer = None):
        return self._handle_decorator(EasyAsNode, name, renderer)
        
    
    def _handle_decorator(self, node_class, name, renderer):
        if not name and not renderer:
            return self.easytag
        if not renderer:
            if callable(name):
                renderer = name
                return self._register_easytag(node_class, renderer.__name__, renderer)
            else:
                def dec(renderer):
                    return self._register_easytag(node_class, name, renderer)
                return dec
        return self._register_easytag(node_class, name, renderer)
    
    def _register_easytag(self, node_class, name, renderer):
        if not renderer:
            renderer = name
            name = renderer.__name__
        
        def render_context(self, context, *args, **kwargs):
            return renderer(context, *args, **kwargs)
        
        get_argspec = classmethod(lambda cls: node_class.get_argspec(renderer))
        
        tag_node = type('%sEasyNode' % name, (node_class,), {
            'render_context': render_context,
            'get_argspec': get_argspec,
        })
        self.tag(name, tag_node.parse)
        return renderer



# -*- coding: utf-8 -*-

'''
Created on 01/03/2011

@author: vbmendes
'''

import unittest

from django import template

from easytags import EasyLibrary
from easytags import EasyNode, EasyAsNode

class LibraryTests(unittest.TestCase):

    def test_easy_library_register_easy_node(self):
        def test_tag(context):
            return u'my return'

        register = EasyLibrary()
        register.easytag(test_tag)
        
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'test_tag')
        
        self.assertTrue(register.tags.has_key('test_tag'))
        
        test_node = register.tags['test_tag'](parser, token)
        
        self.assertTrue(isinstance(test_node, EasyNode))
        
        context = template.Context({})
        
        self.assertEquals(u'my return', test_node.render(context))
    
    def test_easy_library_register_easy_node_with_parameters(self):
        def test_tag(context, arg1):
            return arg1

        register = EasyLibrary()
        register.easytag(test_tag)
        
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'test_tag "my arg"')
        test_node = register.tags['test_tag'](parser, token)
        
        context = template.Context({})
        self.assertEquals(u'my arg', test_node.render(context))
    
    def test_easy_library_register_tags_with_custom_names(self):
        def test_tag(context):
            return u''

        register = EasyLibrary()
        register.easytag('tag_name', test_tag)

        self.assertTrue(register.tags.has_key('tag_name'))
    
    def test_easy_library_register_tags_as_decorating_method(self):
        def test_tag(context):
            return u''

        register = EasyLibrary()
        register.easytag()(test_tag)

        self.assertTrue(register.tags.has_key('test_tag'))
    
    def test_easy_library_register_tags_as_decorating_method_with_name(self):
        def test_tag(context):
            return u''

        register = EasyLibrary()
        register.easytag('tag_name')(test_tag)

        self.assertTrue(register.tags.has_key('tag_name'))
    
    def test_easy_library_register_tags_as_decorating_method_with_name_kwarg(self):
        def test_tag(context):
            return u''

        register = EasyLibrary()
        register.easytag(name='tag_name')(test_tag)

        self.assertTrue(register.tags.has_key('tag_name'))
    
    def test_easy_library_register_tags_keeps_decorated_function_data(self):
        def test_tag(context):
            return u''

        register = EasyLibrary()
        test_tag = register.easytag(name='tag_name')(test_tag)

        self.assertEquals('test_tag', test_tag.__name__)
    
        
    def test_easy_library_register_as_tags(self):
        def test_tag(context):
            return u'my return'

        register = EasyLibrary()
        register.easyastag(test_tag)
        
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'test_tag as varname')
        
        self.assertTrue(register.tags.has_key('test_tag'))
        
        test_node = register.tags['test_tag'](parser, token)
        
        self.assertTrue(isinstance(test_node, EasyAsNode))
        
        context = template.Context({})
        
        self.assertEquals(u'', test_node.render(context))
        self.assertEquals(u'my return', context['varname'])



# -*- coding: utf-8 -*-

'''
Created on 20/02/2011

@author: vbmendes
'''

from inspect import getargspec

from django.template import Node, Variable, TemplateSyntaxError


is_kwarg = lambda bit: not bit[0] in (u'"', u"'") and u'=' in bit


def get_args_kwargs_from_bits(bits):
    args = []
    kwargs = {}
    for bit in bits:
        if is_kwarg(bit):
            splitted_bit = bit.split(u'=')
            kwargs[splitted_bit[0]] = u'='.join(splitted_bit[1:])
        else:
            if not kwargs:
                args.append(bit)
            else:
                raise TemplateSyntaxError(u"Args must be before kwargs.")
                
    return {'args': tuple(args), 'kwargs': kwargs}


class EasyNode(Node):
    
    @classmethod
    def parse_to_args_kwargs(cls, parser, token):
        bits = token.split_contents()
        return get_args_kwargs_from_bits(bits[1:])
    
    @classmethod
    def parse(cls, parser, token):
        args_kwargs = cls.parse_to_args_kwargs(parser, token)
        cls.is_args_kwargs_valid(args_kwargs)
        return cls(args_kwargs)
    
    @classmethod
    def get_argspec(cls, func = None):
        func = func or cls.render_context
        return getargspec(func)
    
    @classmethod
    def is_args_kwargs_valid(cls, args_kwargs):
        render_context_spec = cls.get_argspec()
        
        args = args_kwargs['args']
        kwargs = args_kwargs['kwargs']
        
        valid_args_names = render_context_spec.args
        if 'self' in valid_args_names: valid_args_names.remove('self')
        if 'context' in valid_args_names: valid_args_names.remove('context')
        
        n_args_kwargs = len(args) + len(kwargs)
        
        max_n_args_kwargs = len(valid_args_names) 
        if not render_context_spec.varargs and not render_context_spec.keywords and n_args_kwargs > max_n_args_kwargs:
            raise TemplateSyntaxError(u'Invalid number of args %s (max. %s)' % (n_args_kwargs, max_n_args_kwargs))
        
        min_n_args_kwargs = max_n_args_kwargs - len(render_context_spec.defaults or ())
        if n_args_kwargs < min_n_args_kwargs:
            raise TemplateSyntaxError(u'Invalid number of args %s (min. %s)' % (n_args_kwargs, max_n_args_kwargs))
        
        required_args_names = valid_args_names[len(args):min_n_args_kwargs]
        for required_arg_name in required_args_names:
            if not required_arg_name in kwargs:
                raise TemplateSyntaxError(u'Required arg missing: %s' % required_arg_name)
        
        first_kwarg_index = len(args)
        if not render_context_spec.keywords:
            valid_kwargs = valid_args_names[first_kwarg_index:]
            for kwarg in kwargs:
                if not kwarg in valid_kwargs:
                    raise TemplateSyntaxError(u'Invalid kwarg %s.' % kwarg)
        else:
            defined_args = valid_args_names[:first_kwarg_index]
            for kwarg in kwargs:
                if kwarg in defined_args:
                    raise TemplateSyntaxError(u'%s was defined twice.' % kwarg) 
    
    def __init__(self, args_kwargs):
        self.args = [Variable(arg) for arg in args_kwargs['args']]
        self.kwargs = dict((key, Variable(value)) for key, value in args_kwargs['kwargs'].items())
    
    def render(self, context):
        args = [arg.resolve(context) for arg in self.args]
        kwargs = dict((str(key), value.resolve(context)) for key, value in self.kwargs.items())
        return self.render_context(context, *args, **kwargs)
    
    def render_context(self, context, *args, **kwargs):
        raise NotImplementedError


class EasyAsNode(EasyNode):
    
    @classmethod
    def parse_to_args_kwargs(cls, parser, token):
        bits = token.split_contents()[1:]
        if len(bits) >= 2 and bits[-2] == 'as':
            varname = bits[-1]
            bits = bits[:-2]
        else:
            varname = None
        args_kwargs = get_args_kwargs_from_bits(bits)
        args_kwargs['varname'] = varname
        return args_kwargs
    
    def __init__(self, args_kwargs):
        super(EasyAsNode, self).__init__(args_kwargs)
        self.varname = args_kwargs['varname']
    
    def render(self, context):
        rendered = super(EasyAsNode, self).render(context)
        if self.varname:
            context[self.varname] = rendered
            return u''
        return rendered



import os

from setuptools import setup, find_packages

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "django-meio-easytags",
    version = "0.6",
    url = "http://www.meiocodigo.com/projects/django-meio-easytags/",
    license = "BSD",
    description = "An easy way to create custom template tags for Django's templating system.",
    long_description = read('README.rst'),
    author = "Vinicius Mendes",
    author_email = "vbmendes@gmail.com",
    packages = find_packages('src'),
    package_dir = {'': 'src'},
    install_requires= ['setuptools'],
    classifiers = [
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)




# -*- coding: utf-8 -*-

'''
Created on 20/02/2011

@author: vbmendes
'''

from django import template
from django.template import Context, Variable, TemplateSyntaxError
from django.test import TestCase

from easytags.node import EasyNode, EasyAsNode


class MyEasyNode(EasyNode):

    def render_context(self, context, arg1, kwarg1=None):
        return arg1


class MyEasyNodeWithoutDefaults(EasyNode):

    def render_context(self, context, arg1):
        return arg1

    
class NodeTests(TestCase):
    
    def test_resolves_absolute_string(self):
        context = Context({})
        args_kwargs = {'args': ('"absolute string"',), 'kwargs': {}}
        
        node = MyEasyNode(args_kwargs)
    
        self.assertEquals(
            u'absolute string',
            node.render(context),
        )

    def test_resolve_simple_variable(self):
        context = Context({'simple_variable': u'simple variable value'})
        args_kwargs = {'args': ('simple_variable',), 'kwargs': {}}
        
        node = MyEasyNode(args_kwargs)
    
        self.assertEquals(
            u'simple variable value',
            node.render(context),
        )
    
    def test_resolve_dict_variable(self):
        context = Context({'mydict': {'key': u'value'}})
        args_kwargs = {'args': ('mydict.key',), 'kwargs': {}}
        
        node = MyEasyNode(args_kwargs)
    
        self.assertEquals(
            u'value',
            node.render(context),
        )
    
    def test_resolve_absolute_string_in_kwargs(self):
        context = Context({})
        args_kwargs = {'args': (), 'kwargs': {'arg1': u'"absolute string"'}}
        
        node = MyEasyNode(args_kwargs)
        
        self.assertEquals(
            u'absolute string',
            node.render(context),
        )
    
    def test_resolve_simple_variable_in_kwargs(self):
        context = Context({'simple_variable': u'simple variable value'})
        args_kwargs = {'args': (), 'kwargs': {'arg1': u'simple_variable'}}
        
        node = MyEasyNode(args_kwargs)
        
        self.assertEquals(
            u'simple variable value',
            node.render(context),
        )
    
    def test_resolve_dict_variable_in_kwargs(self):
        context = Context({'mydict': {'key': u'value'}})
        args_kwargs = {'args': (), 'kwargs': {'arg1': 'mydict.key'}}
        
        node = MyEasyNode(args_kwargs)
        
        self.assertEquals(
            u'value',
            node.render(context),
        )

    def test_node_parse_returns_node_instance(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name arg1 kwarg1="a=1"')
        node = MyEasyNode.parse(parser, token)
        
        self.assertTrue(isinstance(node, MyEasyNode))
        self.assertEquals(u'arg1', node.args[0].var)
        self.assertEquals(u'"a=1"', node.kwargs['kwarg1'].var)
    
    def test_node_parse_verifies_invalid_kwarg(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name arg1 invalid_kwarg="a=1"')

        self.assertRaises(TemplateSyntaxError, MyEasyNode.parse, parser, token)
    
    def test_node_parse_verifies_kwarg_already_satisfied_by_arg(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name arg1 arg1="a=1"')

        self.assertRaises(TemplateSyntaxError, MyEasyNode.parse, parser, token)
    
    def test_node_parse_verifies_if_there_are_more_args_kwargs_then_method_requires(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name arg1 arg2 arg3')

        self.assertRaises(TemplateSyntaxError, MyEasyNode.parse, parser, token)

    def test_node_parse_verifies_if_there_are_less_args_kwargs_then_method_requires(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name')

        self.assertRaises(TemplateSyntaxError, MyEasyNode.parse, parser, token)

    def test_node_parse_verifies_if_required_arg_is_specified(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name kwarg1="a"')

        self.assertRaises(TemplateSyntaxError, MyEasyNode.parse, parser, token)

    def test_node_can_have_no_args_with_default_value(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "a"')
        
        node = MyEasyNodeWithoutDefaults.parse(parser, token)
        
        self.assertEquals(u'"a"' ,node.args[0].var)
    
    def test_node_can_receive_infinite_args(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "a" "b" "c" "d"')
        
        MyEasyNode = type('MyEasyNodeWithArgs', (EasyNode,), {
            'render_context': lambda self, context, *args: reduce(lambda x, y: x + y, args)
        })
        
        node = MyEasyNode.parse(parser, token)
        
        self.assertEquals(u'abcd' ,node.render(Context({})))
    
    def test_node_can_receive_required_arg_and_infinite_args(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "a" "b" "c" "d"')
        
        MyEasyNode = type('MyEasyNode', (EasyNode,), {
            'render_context': lambda self, context, arg1, *args: arg1 + reduce(lambda x, y: x + y, args)
        })
        
        node = MyEasyNode.parse(parser, token) 
        
        self.assertEquals(u'abcd' ,node.render(Context({})))
    
    def test_node_verifies_if_required_arg_is_specified_when_node_can_receive_infinite_args(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name')
        
        MyEasyNode = type('MyEasyNode', (EasyNode,), {
            'render_context': lambda self, context, arg1, *args: True
        })
        
        self.assertRaises(TemplateSyntaxError, MyEasyNode.parse, parser, token)

    def test_node_can_receive_kwargs(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name arg1="bla" arg2="ble"')
        
        MyEasyNode = type('MyEasyNode', (EasyNode,), {
            'render_context': lambda self, context, **kwargs:\
                reduce(lambda x,y: u'%s%s' % (x, y),
                    ['%s=%s' % (key, value) for key, value in kwargs.items()])
        })
        
        node = MyEasyNode.parse(parser, token)
        
        self.assertEquals(u'arg1=blaarg2=ble', node.render(Context({})))

    def test_node_verifies_if_required_arg_is_specified_when_code_can_receive_kwargs(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name')
        
        MyEasyNode = type('MyEasyNode', (EasyNode,), {
            'render_context': lambda self, context, arg1, **kwargs: True
        })
        
        self.assertRaises(TemplateSyntaxError, MyEasyNode.parse, parser, token)

    def test_node_verifies_if_required_kwarg_is_specified_when_code_can_receive_kwargs(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name arg2="2"')
        
        MyEasyNode = type('MyEasyNode', (EasyNode,), {
            'render_context': lambda self, context, arg1, **kwargs: True
        })
        
        self.assertRaises(TemplateSyntaxError, MyEasyNode.parse, parser, token)
    
    def test_if_node_can_receive_args_and_kwargs(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "1" arg2="2"')
        
        MyEasyNode = type('MyEasyNode', (EasyNode,), {
            'render_context': lambda self, context, *args, **kwargs: 
                args[0]+kwargs.items()[0][0]+u'='+kwargs.items()[0][1]
        })
        
        node = MyEasyNode.parse(parser, token)
        
        self.assertEquals(u'1arg2=2', node.render(Context({})))
    
    def test_if_node_can_receive_required_arg_and_kwargs(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "required" "2" arg3="3"')
        
        MyEasyNode = type('MyEasyNode', (EasyNode,), {
            'render_context': lambda self, context, arg1, *args, **kwargs: 
                arg1+args[0]+kwargs.items()[0][0]+u'='+kwargs.items()[0][1]
        })
        
        node = MyEasyNode.parse(parser, token)
        
        self.assertEquals(u'required2arg3=3', node.render(Context({})))
    
    def test_node_verifies_if_required_arg_is_specified_two_times(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, 'tag_name "required" arg1="3"')
        
        MyEasyNode = type('MyEasyNode', (EasyNode,), {
            'render_context': lambda self, context, arg1, *args, **kwargs: True
        })
        
        self.assertRaises(TemplateSyntaxError, MyEasyNode.parse, parser, token)
        
    def test_as_node_receives_as_parameter(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, u'tag_name as varname')
        
        MyEasyAsNode = type('MyEasyAsNode', (EasyAsNode,), {
            'render_context': lambda self, context, **kwargs: 'value'
        })
        
        node = MyEasyAsNode.parse(parser, token)
        context = Context()
        
        self.assertEqual('', node.render(context))
        self.assertEqual('value', context['varname']) 
    
    def test_as_node_can_be_used_without_as_parameter(self):
        parser = template.Parser([])
        token = template.Token(template.TOKEN_BLOCK, u'tag_name "value"')
        
        MyEasyAsNode = type('MyEasyAsNode', (EasyAsNode,), {
            'render_context': lambda self, context, arg1, **kwargs: arg1
        })
        
        node = MyEasyAsNode.parse(parser, token)
        context = Context()
        
        self.assertEqual('value', node.render(context))
        
